
#include <X11.h>
#include "defs.h"
#include <intuition/intuitionbas.h>
#include <graphics/gfxbase.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <libraries/dos.h>
#include <stdlib.h>

#define ABS(x)  ((x) < 0 ? -(x) : (x))


Display * sys_display;
int	  sys_screen;
Font	  sys_font;
Cursor	  sys_cursor;
static struct IntuitionBase * IBase = NULL;
struct GfxBase * GfxBase = NULL;
static struct Screen WB;
static struct TextAttr sys_ta;
char * display_name = "";
#ifdef DEC
char * font_name = "terminal_bold10";
#else
char * font_name = "8x13bold";
#endif

#define SYSFONTNAME	font_name

char *ColName[] = {
	"steelblue",
	"white",
	"black",
	"gold",
	NULL
};

long CMap[] = { 0, 0, 0, 0, };

unsigned long plane_mask;


static WIN * FindWindow (rp)
RP * rp;
{
	WIN * w;

	w = (WIN *)WB.FirstWindow;

	while (w) {
		if (w->RPort == rp) break;

		w = (WIN *)w->NextWindow;
	}

	return (w);
}

void SetAPen (rp, col)
RP * rp;
long col;
{
	if (col == -1) col = 1;
	col &= 3;

	XSetForeground (sys_display, rp->gc, CMap[col]);

	rp->FgPen = col;
}

void SetBPen (rp, col)
RP * rp;
long col;
{
	if (col == -1) col = 0;
	col &= 3;

	XSetBackground (sys_display, rp->gc, CMap[col]);

	rp->BgPen = col;
}

void SetDrMd (rp, mode)
RP * rp;
int mode;
{
	if (mode & COMPLEMENT) {
		XSetFunction (sys_display, rp->gc, GXxor);
	} else {
		XSetFunction (sys_display, rp->gc, GXcopy);
	}

	rp->DrawMode = mode;
}

void RectFill (rp, x1, y1, x2, y2)
RP * rp;
long x1, y1, x2, y2;
{
	WIN * w;

	if (!(w = FindWindow(rp))) return;

	XFillRectangle (sys_display, w->xwindow, rp->gc, x1, y1,
		x2-x1+1, y2-y1+1);
}

#define SWAP(a,b)       { a ^= b; b ^= a; a ^= b; }

void ScrollRaster (rp, dx, dy, x1, y1, x2, y2)
RP * rp;
long dx, dy, x1, y1, x2, y2;
{
	long w, h, x3, y3, x4, y4, _dx_, _dy_;
	long apen = rp->FgPen;
	WIN * win;

	if (!dx && !dy) return;

	if (!(win = FindWindow(rp))) return;

	if (x2 < x1) SWAP (x1,x2)
	if (y2 < y1) SWAP (y1,y2)

	_dx_ = ABS(dx);
	_dy_ = ABS(dy);

	x3 = x1 + _dx_;
	y3 = y1 + _dy_;

	x4 = x2 - _dx_ +1;
	y4 = y2 - _dy_ +1;

	w = x2 - x3 +1;
	h = y2 - y3 +1;

	SetAPen (rp, rp->BgPen);

	if (dx <= 0) {
		if (dy <= 0) {
			XCopyArea (sys_display,win->xwindow,win->xwindow,rp->gc,
				x1, y1, w, h, x3, y3);

			if (_dy_) XClearArea (sys_display,win->xwindow,
				x1, y1, w+_dx_, _dy_, FALSE);

			if (_dx_) XClearArea (sys_display,win->xwindow,
				x1, y1, _dx_, h, FALSE);

		} else { /* dy > 0 */
			XCopyArea (sys_display,win->xwindow,win->xwindow,rp->gc,
				x1, y3, w, h, x3, y1);

			XClearArea (sys_display,win->xwindow,
				x1, y4, w+_dx_, _dy_, FALSE);

			if (_dx_) XClearArea (sys_display,win->xwindow,
				x1, y1, _dx_, h, FALSE);
		}
	} else { /* dx > 0 */
		if (dy <= 0) {
			XCopyArea (sys_display,win->xwindow,win->xwindow,rp->gc,
				x3, y1, w, h, x1, y3);

			if (_dy_) XClearArea (sys_display,win->xwindow,
				x1, y1, w+_dx_, _dy_, FALSE);

			XClearArea (sys_display,win->xwindow,
				x4, y3, _dx_, h, FALSE);
		} else { /* dy > 0 */
			XCopyArea (sys_display,win->xwindow,win->xwindow,rp->gc,
				x3, y3, w, h, x1, y1);

			XClearArea (sys_display,win->xwindow,
				x1, y4, w+_dx_, _dy_, FALSE);

			XClearArea (sys_display,win->xwindow,
				x4, y1, _dx_, h, FALSE);
		}
	}

	SetAPen (rp, apen);
}

void * AllocMem (size, reqs)
long size, reqs;
{
	void * memory;

	memory = malloc(size);

	if (memory) {
		if (reqs & MEMF_CLEAR)
			memset (memory, 0, size);
	}

	return (memory);
}

void FreeMem (ptr, size)
void * ptr;
long size;
{
	free (ptr);
}

void Text (rp, string, len)
RP * rp;
char * string;
long len;
{
	WIN * w;

	if (!(w = FindWindow(rp))) return;

	if (rp->DrawMode & JAM2)
		XDrawImageString (sys_display, w->xwindow, rp->gc, rp->cp_x,
			rp->cp_y, string, len);
	else	XDrawString (sys_display, w->xwindow, rp->gc, rp->cp_x,
			rp->cp_y, string, len);
}

void Move (rp, x, y)
RP * rp;
long x, y;
{
	rp->cp_x = x;
	rp->cp_y = y;
}

void Draw (rp, x, y)
RP * rp;
long x, y;
{
	WIN * w;

	if (!(w = FindWindow(rp))) return;

	XDrawLine (sys_display, w->xwindow, rp->gc, rp->cp_x, rp->cp_y,
		x, y);
	rp->cp_x = x;
	rp->cp_y = y;
}

void SetRast (rp, color)
RP * rp;
long color;
{
	WIN * w;

	if (!(w = FindWindow(rp))) return;

	XClearArea (sys_display, w->xwindow, w->BorderLeft, w->BorderTop,
		w->Width - w->BorderLeft - w->BorderRight,
		w->Height - w->BorderTop - w->BorderBottom, FALSE);
}

void WindowToFront (window)
WIN * window;
{
	XRaiseWindow (sys_display, window->xwindow);
}

void SetFont (rp, font)
RP * rp;
FONT * font;
{
	if (rp->gc)
	    XSetFont (sys_display, rp->gc, font->xfs.fid);

	rp->Font = font;
	rp->TxWidth = font->tf_XSize;
	rp->TxHeight = font->tf_YSize;
	rp->TxBaseline = font->tf_Baseline;
	font->tf_Accessors ++;
}

FONT * OpenFont (ta)
struct TextAttr * ta;
{
	FONT * tf;
	XFontStruct * xfs;

	tf = (FONT *)malloc (sizeof(FONT));

	if (!tf) return (NULL);

	xfs = XLoadQueryFont (sys_display, ta->ta_Name);

	if (!xfs) {
		free (tf);
		return (NULL);
	}

	tf->xfs = *xfs;

	tf->tf_YSize = tf->xfs.max_bounds.ascent +
			tf->xfs.max_bounds.descent;
	tf->tf_XSize = tf->xfs.max_bounds.rbearing -
			tf->xfs.min_bounds.lbearing;
	tf->tf_Baseline = tf->xfs.ascent;
	tf->tf_LoChar = tf->xfs.min_char_or_byte2;
	tf->tf_HiChar = tf->xfs.max_char_or_byte2;
	tf->tf_Accessors ++;

	if (!tf->tf_XSize || !tf->tf_YSize) {
		fprintf ("Error: Fontsize");
		exit (1);
	}

	return (tf);
}

FONT * OpenDiskFont (ta)
struct TextAttr * ta;
{
	return (OpenFont(ta));
}

void CloseFont (tf)
FONT * tf;
{
	tf->tf_Accessors --;
	if (!tf->tf_Accessors) {
	    XUnloadFont (sys_display, tf->xfs.fid);
	    free (tf);
	}
}

long DeleteFile (name)
char * name;
{
	return (unlink(name));
}

void RefreshWindowFrame (w)
WIN * w;
{
	return;
}

void * OpenLibrary (name, ver)
char * name;
int ver;
{
	void * libbase;

	init_unix ();

	if (strcmp(name, "intuition.library") == 0) {
		IBase = AllocMem (sizeof(struct IntuitionBase), MEMF_CLEAR);
		libbase = IBase;
		memset (&WB, 0, sizeof(struct Screen));
		IBase->FirstScreen = IBase->ActiveScreen = &WB;
		WB.Width = DisplayWidth (sys_display, sys_screen);
		WB.Height = DisplayHeight (sys_display, sys_screen);
	} else if (strcmp(name, "graphics.library") == 0) {
		FONT * def;

		GfxBase = AllocMem (sizeof(struct GfxBase), MEMF_CLEAR);
		libbase = GfxBase;

		sys_ta.ta_Name = (STRPTR)SYSFONTNAME;
		def = OpenFont (&sys_ta);

		if (!def) {
		    fprintf (stderr, "Cannot open font %s; trying fixed\n",
			    SYSFONTNAME);

		    sys_ta.ta_Name = (STRPTR)"fixed";
		    def = OpenFont (&sys_ta);

		    if (!def) {
			    fprintf (stderr, "Cannot open font\n");
			    return (NULL);
		    }
		}

		GfxBase->DefaultFont = (struct TextFont *)def;
		def->tf_Accessors ++;

		SetFont (&WB.RastPort, def);
		WB.Font = &sys_ta;
		sys_ta.ta_YSize = def->tf_YSize;
	}

	return (libbase);
}

void Forbid()
{
	return;
}

void Permit ()
{
	return;
}

void CloseLibrary (ptr)
void * ptr;
{
	free (ptr);
}

int IntuiTextLength (itext)
struct IntuiText * itext;
{
	return (100);
}

long Execute (cmd, in, out)
char * cmd;
int in;
int out;
{
	return (system (cmd));
}

void SizeWindow (win, dx, dy)
WIN * win;
int dx, dy;
{
	XResizeWindow (sys_display, win->xwindow, win->Width+dx,
		win->Height + dy);
}

void Delay (time)
int time;
{
	time /= 50;
	if (time) sleep (time);
}

void ClearMenuStrip (win)
WIN * win;
{
	return;
}

void ActivateWindow (win)
WIN * win;
{
	XSetInputFocus (sys_display, win->xwindow, RevertToNone, CurrentTime);

	IBase->ActiveWindow = (struct Window *)win;
}

void CloseWindow (win)
WIN * win;
{
	WIN * wptr;

	wptr = (WIN *)WB.FirstWindow;

	if (win == wptr) {
	    wptr = (WIN *)WB.FirstWindow = win->NextWindow;
	} else {
	    while (wptr->NextWindow != (struct Window *)win)
		wptr = (WIN *)wptr->NextWindow;

	    if (!wptr)
		return;

	    wptr->NextWindow = win->NextWindow;
	}

	IBase->ActiveWindow = (struct Window *)wptr;

	XFreeGC (sys_display, win->RPort->gc);
	XDestroyWindow (sys_display, win->xwindow);

	free (win->RPort);

	if (win->UserPort)
	    free (win->UserPort);

	free (win);
}

void DeletePort (port)
struct MsgPort * port;
{
	return ;
}

void AddHead (list, node)
struct List * list;
struct Node * node;
{
	list->lh_Head->ln_Pred = node;
	node->ln_Succ = list->lh_Head;
	list->lh_Head = node;
	node->ln_Pred = (struct Node *)&list->lh_Head;
}

void AddTail (list, node)
struct List * list;
struct Node * node;
{
	node->ln_Succ = (struct Node *)list->lh_Tail;
	list->lh_TailPred->ln_Succ = node;
	node->ln_Pred = list->lh_TailPred;
	list->lh_TailPred = node;
}

void NewList (list)
struct List * list;
{
	list->lh_Head = (struct Node *)&list->lh_Tail;
	list->lh_Tail = NULL;
	list->lh_TailPred = (struct Node *)list;
}

void Remove (node)
struct Node * node;
{
	struct Node * succ, * pred;

	succ = node->ln_Succ;
	pred = node->ln_Pred;
	pred->ln_Succ = succ;
	succ->ln_Pred = pred;
}

void ReplyMsg (msg)
struct Message * msg;
{
	free (msg);
}

extern void ModifyIDCMP ();

WIN * OpenWindow (nw)
struct NewWindow * nw;
{
	WIN * w, * parent;
	RP * rp;
	XGCValues gcval;
	void SetWindowTitles();

	w = AllocMem (sizeof(WIN), MEMF_CLEAR);
	rp = AllocMem (sizeof(RP), MEMF_CLEAR);
	if (nw->IDCMPFlags)
	    w->UserPort = AllocMem (sizeof(struct MsgPort), MEMF_CLEAR);

	w->LeftEdge = nw->LeftEdge;
	w->TopEdge = nw->TopEdge;
	w->Width = nw->Width;
	w->Height = nw->Height;
	w->RPort = rp;
	w->IDCMPFlags = nw->IDCMPFlags;

	if (nw->DetailPen == 0xff) nw->DetailPen = 1;
	if (nw->BlockPen == 0xff) nw->BlockPen = 0;

	w->xwindow = XCreateSimpleWindow (sys_display,
		DefaultRootWindow(sys_display),
		nw->LeftEdge, nw->TopEdge, nw->Width, nw->Height, 15,
		CMap[1], CMap[0]);

	SetWindowTitles (w, nw->Title, -1);

	if (w->UserPort)
	    w->UserPort->win = (APTR)w;

	gcval.plane_mask = plane_mask;

	rp->gc = XCreateGC (sys_display, w->xwindow, GCPlaneMask, &gcval);
	w->BorderLeft = 2;
	w->BorderTop = GfxBase->DefaultFont->tf_YSize + 3;
	w->BorderRight = 2;
	w->BorderBottom = 2;

	w->WScreen = &WB;

	SetFont (rp, GfxBase->DefaultFont);

	SetAPen (rp, nw->DetailPen);
	SetBPen (rp, nw->BlockPen);
	SetDrMd (rp, JAM2);

	if (WB.FirstWindow) {
		parent = (WIN *)WB.FirstWindow;
		while (parent->NextWindow)
			parent = (WIN *)parent->NextWindow;

		parent->NextWindow = (struct Window *)w;
	} else {
		WB.FirstWindow = (struct Window *)w;
	}

	w->NextWindow = NULL;

	if (nw->Flags & ACTIVATE) IBase->ActiveWindow = (struct Window *)w;

	XSelectInput (sys_display, w->xwindow, ExposureMask |
	    ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
	    KeyPressMask | KeyReleaseMask |
	    EnterWindowMask | LeaveWindowMask);

	ModifyIDCMP (w, w->IDCMPFlags);

	XMapRaised (sys_display, w->xwindow);
	XDefineCursor (sys_display, w->xwindow, sys_cursor);

	/*ActivateWindow (IBase->ActiveWindow);*/

	return (w);
}

void ModifyIDCMP (win, flags)
WIN * win;
long flags;
{
	if (!win->IDCMPFlags && flags) {
	    win->UserPort = AllocMem (sizeof(struct MsgPort), MEMF_CLEAR);
	    win->UserPort->win = (APTR)win;
	}

	win->IDCMPFlags = flags;

	if (!flags) {
	    if (win->UserPort) {
		free (win->UserPort);
		win->UserPort = NULL;
	    }
	}
}

long OpenDevice (devname, unit, ioreq, flags)
char * devname;
long unit;
struct IORequest * ioreq;
long flags;
{
	return (0);
}

long Wait ()
{
	XEvent event;

/*	XPeekEvent (sys_display, &event);*/

	return (-1L);
}

struct _keytable {
	KeySym keysym;
	WORD amiga;
	UWORD amiga_qual;
	char * normal, * shifted;
	ULONG keycode;
} keytable[] = {
{XK_Return, 0x44, 0, "\012", "\012", 0 },
{XK_Right, 0x4e, 0, "\233C", "\233 A", 0 },
{XK_Up, 0x4c, 0, "\233A", "\233T",0 },
{XK_Left, 0x4f, 0, "\233D", "\233 @",0 },
{XK_Down, 0x4d, 0, "\233B", "\233S",0 },
{XK_Help, 0x5f, 0, "\233?~", "\233?~",0 },
{XK_KP_Enter, 0x43, IEQUALIFIER_NUMERICPAD, "\015", "\015",0 },
{XK_KP_Separator, 0x3c, IEQUALIFIER_NUMERICPAD, ".", ".",0 },
{XK_KP_Subtract, 0x4a, IEQUALIFIER_NUMERICPAD, "-", "-",0 },
{XK_KP_Decimal, 0x3c, IEQUALIFIER_NUMERICPAD, ".", ".",0 },
{XK_KP_0, 0x0f, IEQUALIFIER_NUMERICPAD, "0", "0",0 },
{XK_KP_1, 0x1d, IEQUALIFIER_NUMERICPAD, "1", "1",0 },
{XK_KP_2, 0x1e, IEQUALIFIER_NUMERICPAD, "2", "2",0 },
{XK_KP_3, 0x1f, IEQUALIFIER_NUMERICPAD, "3", "3",0 },
{XK_KP_4, 0x2d, IEQUALIFIER_NUMERICPAD, "4", "4",0 },
{XK_KP_5, 0x2e, IEQUALIFIER_NUMERICPAD, "5", "5",0 },
{XK_KP_6, 0x2f, IEQUALIFIER_NUMERICPAD, "6", "6",0 },
{XK_KP_7, 0x3d, IEQUALIFIER_NUMERICPAD, "7", "7",0 },
{XK_KP_8, 0x3e, IEQUALIFIER_NUMERICPAD, "8", "8",0 },
{XK_KP_9, 0x3f, IEQUALIFIER_NUMERICPAD, "9", "9",0 },
{XK_F1, 0x50, 0, "\2330~", "\23310~",0 },
{XK_F2, 0x51, 0, "\2331~", "\23311~",0 },
{XK_F3, 0x52, 0, "\2332~", "\23312~",0 },
{XK_F4, 0x53, 0, "\2333~", "\23313~",0 },
{XK_F5, 0x54, 0, "\2334~", "\23314~",0 },
{XK_F6, 0x55, 0, "\2335~", "\23315~",0 },
{XK_F7, 0x56, 0, "\2336~", "\23316~",0 },
{XK_F8, 0x57, 0, "\2337~", "\23317~",0 },
{XK_F9, 0x58, 0, "\2338~", "\23318~",0 },
{XK_F10, 0x59, 0, "\2339~", "\23319~",0 },
{0, -1, 0, },
};

#define SHIFT	(IEQUALIFIER_LSHIFT | IEQUALIFIER_RSHIFT)
#define ALT	(IEQUALIFIER_LALT | IEQUALIFIER_RALT)
#define CTRL	IEQUALIFIER_CONTROL
#define CAPS	IEQUALIFIER_CAPSLOCK

long StateToQualifier (state)
unsigned long state;
{
	long result;

	result = 0;

	if (state & ShiftMask)
		result |= SHIFT;

	if (state & ControlMask)
		result |= CTRL;

	if (state & LockMask)
		result |= CAPS;

	if (state & Mod1Mask)
		result |= ALT;

	if (state & Mod2Mask)
		result |= AMIGAKEYS;

	if (state & Button1Mask)
		result |= IEQUALIFIER_LEFTBUTTON;

	if (state & Button2Mask)
		result |= IEQUALIFIER_RBUTTON;

	if (state & Button3Mask)
		result |= IEQUALIFIER_MIDBUTTON;

	return (result);
}

long XKeyToAmigaCode (xk)
XKeyEvent * xk;
{
	char buffer[10];
	KeySym ks;
	int count;
	long result;
	short t;

	result = 0;

	if (xk->state & ShiftMask)
		result |= SHIFT;

	if (xk->state & ControlMask)
		result |= CTRL;

	if (xk->state & LockMask)
		result |= CAPS;

	if (xk->state & Mod1Mask)
		result |= ALT;

	if (xk->state & Mod2Mask)
		result |= AMIGAKEYS;

	if (xk->state & Button1Mask)
		result |= IEQUALIFIER_LEFTBUTTON;

	if (xk->state & Button2Mask)
		result |= IEQUALIFIER_RBUTTON;

	if (xk->state & Button3Mask)
		result |= IEQUALIFIER_MIDBUTTON;

	result <<= 16;

	xk->state = 0;
	count = XLookupString (xk, buffer, 10, &ks, NULL);

	for (t=0; keytable[t].amiga != -1; t++) {
		if (ks == keytable[t].keycode) {
			result |= (keytable[t].amiga_qual << 16) |
					keytable[t].amiga;
			return (result);
		}
	}

	result |= xk->keycode & 0xffff;

	return (result);
}

void * GetMsg (port)
struct MsgPort * port;
{
	IMESS * im;
	WIN * w;
	char * ptr;

	im = AllocMem (sizeof(IMESS), MEMF_CLEAR);

next:	XNextEvent (sys_display, &(im->event));

	if (im->event.type == MappingNotify) {
		XRefreshKeyboardMapping ((XMappingEvent*)&im->event);
		free (im);
		return (NULL);
	}

	w = (WIN *)WB.FirstWindow;

	while (w) {
		if (w->xwindow == im->event.xany.window) break;

		w = (WIN *)w->NextWindow;
	}

	if (!w) goto next;

	im->IDCMPWindow = (struct Window *)w;

	switch (im->event.type) {
		case Expose: {
			XWindowAttributes win_attr;

			if (im->event.xexpose.count != 0) break;

			XGetWindowAttributes (sys_display,
				w->xwindow, &win_attr);

			if (win_attr.width != w->Width ||
				win_attr.height != w->Height) {
				w->Width = win_attr.width;
				w->Height = win_attr.height;
				im->Class = NEWSIZE;
				ptr = "NEWSIZE";
			} else {
				im->Class = REFRESHWINDOW;
				ptr = "REFRESHWINDOW";
			}
		} break;

		case ButtonPress: {
			XButtonEvent * xb = &im->event.xbutton;

			im->Class = MOUSEBUTTONS;
			im->Qualifier = StateToQualifier (xb->state);
			im->MouseX = xb->x;
			im->MouseY = xb->y;

			switch (xb->button) {
				case Button1:
					im->Code = SELECTDOWN;
					break;

				case Button2:
					im->Code = MIDDOWN;
					break;

				case Button3:
					im->Code = MENUDOWN;
					break;
			}

			ptr = "MOUSEBUTTONS";
		} break;

		case ButtonRelease: {
			XButtonEvent * xb = &im->event.xbutton;

			im->Class = MOUSEBUTTONS;
			im->Qualifier = StateToQualifier (xb->state);
			im->MouseX = xb->x;
			im->MouseY = xb->y;

			switch (xb->button) {
				case Button1:
					im->Code = SELECTUP;
					break;

				case Button2:
					im->Code = MIDUP;
					break;

				case Button3:
					im->Code = MENUUP;
					break;
			}

			ptr = "MOUSEBUTTONS";
		} break;

		case KeyPress: {
			XKeyEvent * xk = &im->event.xkey;
			ULONG result;

			im->Class = RAWKEY;
			result = XKeyToAmigaCode(xk);
			im->Code = xk->keycode;
			im->Qualifier = result >> 16;

			ptr = NULL;
		} break;

		case KeyRelease: {
			XKeyEvent * xk = &im->event.xkey;
			ULONG result;

			im->Class = RAWKEY;
			result = XKeyToAmigaCode(xk);
			im->Code = xk->keycode | 0x8000;
			im->Qualifier = result >> 16;

			ptr = NULL;
		} break;

		case MotionNotify: {
			XMotionEvent * xm = &im->event.xmotion;

			im->Code = QMOVE;
			im->Class = MOUSEMOVE;
			im->Qualifier = StateToQualifier (xm->state);
			im->MouseX = xm->x;
			im->MouseY = xm->y;

			ptr = "MOUSEMOVE";
		} break;

		case EnterNotify: {
			XCrossingEvent * xc = &im->event.xcrossing;

			im->Class = ACTIVEWINDOW;
			im->MouseX = xc->x;
			im->MouseY = xc->y;

			ptr = "ACTIVEWINDOW";
		} break;

		case LeaveNotify: {
			XCrossingEvent * xc = &im->event.xcrossing;

			im->Class = ACTIVEWINDOW;
			im->MouseX = xc->x;
			im->MouseY = xc->y;

			ptr = "INACTIVEWINDOW";
		} break;

		default:
			free (im);
			return (NULL);
		break;
	}

	if (!(im->Class & w->IDCMPFlags) ) {
		free (im);
		return (NULL);
	}

	/*if (ptr)
	printf ("Message : %s  Code %3d %02x  Qualifier %4x\n", ptr,
		im->Code, im->Code, im->Qualifier);*/

	return ((void *)im);
}

int AutoRequest (win, body, pos, neg, f_pos, f_neg, width, height)
WIN * win;
struct IntuiText * body, * pos, * neg;
ULONG f_pos, f_neg;
SHORT width, height;
{
	return (TRUE);
}

void SetWindowTitles (win, text, screen)
WIN * win;
char * text, * screen;
{
	XSizeHints hints;

	hints.x = win->LeftEdge;
	hints.y = win->TopEdge;
	hints.width = win->Width;
	hints.height = win->Height;
	hints.flags = PPosition | PSize;

	if (screen == (char *)-1) screen = "Amiga";

	XSetStandardProperties (sys_display, win->xwindow, text, screen,
		None, NULL, NULL, &hints);
}

void GetScreenData (scr, size, type, screen)
struct Screen * scr, * screen;
long size, type;
{
	if (type != WBENCHSCREEN) {
		movmem (scr, screen, size);
	} else {
		WB.Width = DisplayWidth(sys_display, sys_screen);
		WB.Height = DisplayHeight (sys_display, sys_screen);
		movmem (scr, &WB, size);
	}
}

int RawKeyConvert (ie, buf, size, km)
IEV * ie;
char * buf;
long size;
struct KeyMap * km;
{
	XKeyEvent xk;
	char * ptr;
	int t;

	ie->ie_Code &= 0x7fff;

	for (t=0; keytable[t].amiga != -1; t++) {
		if (ie->ie_Code == keytable[t].keycode) {
			if (ie->ie_Qualifier & SHIFT)
				ptr = keytable[t].shifted;
			else	ptr = keytable[t].normal;

			t = strlen(ptr);
			if (t > size) t = size;

			strncpy (buf, ptr, t);

			goto ende;
		}
	}

	xk.keycode = ie->ie_Code;
	xk.display = sys_display;
	xk.state = 0;

	if (ie->ie_Qualifier & SHIFT)
		xk.state |= ShiftMask;

	if (ie->ie_Qualifier & CTRL)
		xk.state |= ControlMask;

	if (ie->ie_Qualifier & CAPS)
		xk.state |= LockMask;

	if (ie->ie_Qualifier & ALT)
		xk.state |= Mod1Mask;

	if (ie->ie_Qualifier & AMIGAKEYS)
		xk.state |= Mod2Mask;

	if (ie->ie_Qualifier & IEQUALIFIER_LEFTBUTTON)
		xk.state |= Button1Mask;

	if (ie->ie_Qualifier & IEQUALIFIER_MIDBUTTON)
		xk.state |= Button2Mask;

	if (ie->ie_Qualifier & IEQUALIFIER_RBUTTON)
		xk.state |= Button3Mask;

	t = XLookupString(&xk, buf, size, NULL, NULL);

	if (!*buf && t == 1) t = 0;
	if (!t) *buf = 0;

ende:	/*printf ("RawKeyConvert: In %02x %04x %04x Out : %d cs %02x '%c'\n",
		ie->ie_Code, ie->ie_Qualifier, xk.state, t, (ubyte)*buf,
		(ubyte)*buf);*/

	return (t);
}

long TaskFlags = 0;

long SetSignal (bits, mask)
long bits, mask;
{
	long ret= TaskFlags;
	long set, reset;

	set = bits & mask;
	reset = mask & ~set;

	TaskFlags |= set;
	TaskFlags &= ~reset;

	return (ret);
}

struct _Lock {
	char pfad[256];
	mode_t old_mode;
	int access;
};

BPTR CurrentDir (lock)
BPTR lock;
{
	struct _Lock * old;
	struct _Lock * new;
	struct stat file_stat;

	new = (struct _Lock *)lock;

	old = (struct _Lock *)malloc(sizeof(struct _Lock));

	getcwd (old->pfad, 256);
	if (stat(old->pfad, &file_stat))
	    return (NULL);

	old->old_mode = file_stat.st_mode;
	old->access = 0;

	if (new)
	    chdir (new->pfad);

	return ((BPTR)old);
}

BPTR Lock (name, mode)
char * name;
long mode;
{
	struct _Lock * new;
	struct stat file_stat;
	mode_t new_mode;

	if (stat(name, &file_stat))
	    return (NULL);

	new = (struct _Lock *)malloc(sizeof(struct _Lock));

	strcpy (new->pfad, name);
	new->old_mode = file_stat.st_mode;
	new->access = 0;

	new_mode = new->old_mode & S_IRWXU;
	if (mode == SHARED_LOCK)
	    new_mode |= new->old_mode & (S_IRGRP|S_IROTH);

	chmod (name, new_mode);

	return ((BPTR)new);
}

void UnLock (lock)
struct _Lock * lock;
{
	if (lock->access) {
	    lock->access --;
	    return;
	}

	chmod (lock->pfad, lock->old_mode);

	free (lock);
}

BPTR DupLock (lock)
struct _Lock * lock;
{
	lock->access ++;

	return ((BPTR)lock);
}

struct Process * myself = NULL;

struct Task * FindTask (name)
char * name;
{
	if (!myself) {
		BPTR lock = CurrentDir (NULL);

		myself = (struct Process *)malloc (sizeof(struct Process));

		myself->pr_CurrentDir = lock;
		myself->pr_CIS = (BPTR)stdin;
		myself->pr_COS = (BPTR)stdout;
	}

	return ((struct Task *)myself);
}

BPTR ParentDir (lock)
struct _Lock * lock;
{
	char * ptr;
	struct _Lock * new = (struct _Lock *)DupLock (lock);

	ptr = new->pfad + strlen(new->pfad);

	if (*ptr == '/') ptr --;

	while (*ptr != '/' && ptr != new->pfad) ptr --;

	if (ptr != new->pfad) {
		free (new);
		return (NULL);
	}

	*ptr = 0;

	return ((BPTR)new);
}

long Examine (lock, fib)
struct _Lock * lock;
struct FileInfoBlock * fib;
{
	struct stat stat_buf;
	char name[256], * ptr;

	if (stat(lock->pfad, &stat_buf) == -1) return 0;

	fib->fib_DiskKey = stat_buf.st_ino;
	if (stat_buf.st_mode & S_IFDIR)
		fib->fib_DirEntryType = 1;
	else	fib->fib_DirEntryType = -1;

	ptr = lock->pfad;

	ptr += strlen(ptr);

	if (*ptr == '/') *ptr = 0;

	while (*ptr != '/' && ptr != lock->pfad) ptr --;

	if (*ptr == '/') ptr ++;

	strcpy (fib->fib_FileName, ptr);

	fib->fib_Protection = 0;

	if (stat_buf.st_mode & S_IWRITE)
		fib->fib_Protection |= FIBF_WRITE | FIBF_DELETE;

	if (stat_buf.st_mode & S_IREAD)
		fib->fib_Protection |= FIBF_READ;

	if (stat_buf.st_mode & S_IEXEC)
		fib->fib_Protection |= FIBF_EXECUTE;

	fib->fib_Size = stat_buf.st_size;
	fib->fib_NumBlocks = (stat_buf.st_size + 511) / 512;

	return (1);
}

BPTR Open (name, mode)
char * name;
long mode;
{
	FILE * fd;

	switch (mode) {
		case MODE_OLDFILE:
			fd = fopen (name, "a");

			if (!fd) return (NULL);

			rewind (fd);

			break;

		case MODE_NEWFILE:
			fd = fopen (name, "w");

			break;
	}

	return ((BPTR)fd);
}

void Close (fd)
FILE * fd;
{
	fclose (fd);
}

BPTR Output ()
{
	struct Process * p;

	p = (struct Process *)FindTask(NULL);

	return (p->pr_COS);
}

long Write (fd, buf, size)
FILE * fd;
char * buf;
int size;
{
	return (fwrite(buf, size, 1, fd));
}

#define INIT		char * ptr = instring;
#define GETC()          (*ptr++)
#define PEEKC()         (*ptr)
#define UNGETC(c)       (--ptr);
#define RETURN(c)       return;
#define ERROR(c)

#include <regexp.h>

int WildCmp (wild, test)
char * wild, * test;
{
	char expbuf[256];

	compile (wild, expbuf, &expbuf[255], '\0');

	if (step(test, expbuf))
		return (1);

	return (0);
}

init_unix ()
{
	Colormap cm;
	XColor xc;
	XColor fg, bg;
	short t;
	short depth;

	if (sys_display) return;

	sys_display = XOpenDisplay (display_name);

	if (!sys_display) {
	    sys_display = XOpenDisplay (getenv("DISPLAY"));
	    if (!sys_display) {
		printf ("Kein X-Server vorhanden (%s)!\n", display_name);
		exit (1);
	    }
	}

	sys_screen = DefaultScreen (sys_display);

	depth = DisplayPlanes (sys_display, sys_screen);
	cm = DefaultColormap (sys_display, sys_screen);

	plane_mask = 0;

	for (t=0; ColName[t]; t++) {
		if (depth == 1) {
			if (t & 1)
				CMap[t] = WhitePixel(sys_display, sys_screen);
			else	CMap[t] = BlackPixel(sys_display, sys_screen);
		} else {
		    if (XParseColor (sys_display, cm, ColName[t], &xc)) {
			if (!XAllocColor (sys_display, cm, &xc)) {
				printf ("Coulnd't allocate color %s\n",
					ColName[t]);
				exit (1);
			}

			CMap[t] = xc.pixel;

			if (t == 0)
			    bg = xc;
			else if (t == 1)
			    fg = xc;
		    } else {
			printf ("Couldn't get color %s\n", ColName[t]);
			exit (1);
		    }
		}

		plane_mask |= CMap[t];
	}

	sys_cursor = XCreateFontCursor (sys_display, XC_xterm);
	XRecolorCursor (sys_display, sys_cursor, &fg, &bg);

	for (t=0; keytable[t].amiga != -1; t++)
		keytable[t].keycode = XKeysymToKeycode (sys_display,
			keytable[t].keysym);
}

